/****************************************************
   SERIAL.C  Unit fr V24-Ansteuerung fr LPC1751
 ****************************************************/

#include "StdTypes.h"
#include "LPC17xx.h"
#include "startup.h"
#include "config.h"
#include "gio.h"
#include "serial.h"

/* Der Empfang verlt sich auf den Hardware-Fifo
   und das Senden erfolgt gepuffert ber V24puffer per Interrupt.

   Beim Senden wird eine begrenzte Zeit gewartet, wenn der Sendepuffer fast voll ist
   Beim Empfangen wird nicht gewartet, sondern 0 geliefert, wenn nix anliegt.
   Also vorher prfen, ob Zeichen im Rx anliegen
 */

#define v24bufsize 256          /* immer 2er Potenz       */

char V24puffer0[v24bufsize];    /* Sendepuffer */
volatile int v24sin0;           /* Index im Sendepuffer 0 */
volatile int v24sout0;          /* Index im Sendepuffer 0 */

char V24puffer1[v24bufsize];    /* Sendepuffer */
volatile int v24sin1;           /* Index im Sendepuffer 1 */
volatile int v24sout1;          /* Index im Sendepuffer 1 */

char V24puffer2[v24bufsize];    /* Sendepuffer */
volatile int v24sin2;           /* Index im Sendepuffer 2 */
volatile int v24sout2;          /* Index im Sendepuffer 2 */

char V24puffer3[v24bufsize];    /* Sendepuffer */
volatile int v24sin3;           /* Index im Sendepuffer 3 */
volatile int v24sout3;          /* Index im Sendepuffer 3 */



struct FracEntry
{ byte div;    /* Faktor*16 - 256        */
  byte FDR;    /* (MULVAL<<4) | DIVADVAL */
};

/* Hilfstafel: (K * 16) - 256, Code fr UxFDR,   Teilerverhltnis */
const struct FracEntry FracTabelle[72] =
{{   0, 0x10 },  /*  0.  1.000 */
 {  17, 0xF1 },  /*  1.  1.067 */
 {  18, 0xE1 },  /*  2.  1.071 */
 {  20, 0xD1 },  /*  3.  1.077 */
 {  21, 0xC1 },  /*  4.  1.083 */
 {  23, 0xB1 },  /*  5.  1.091 */
 {  26, 0xA1 },  /*  6.  1.100 */
 {  28, 0x91 },  /*  7.  1.111 */
 {  32, 0x81 },  /*  8.  1.125 */
 {  34, 0xF2 },  /*  9.  1.133 */
 {  37, 0x71 },  /* 10.  1.143 */
 {  39, 0xD2 },  /* 11.  1.154 */
 {  43, 0x61 },  /* 12.  1.167 */
 {  47, 0xB2 },  /* 13.  1.182 */
 {  51, 0x51 },  /* 14.  1.200 */
 {  55, 0xE3 },  /* 15.  1.214 */
 {  57, 0x92 },  /* 16.  1.222 */
 {  59, 0xD3 },  /* 17.  1.231 */
 {  64, 0x41 },  /* 18.  1.250 */
 {  68, 0xF4 },  /* 19.  1.267 */
 {  70, 0xB3 },  /* 20.  1.273 */
 {  73, 0x72 },  /* 21.  1.286 */
 {  77, 0xA3 },  /* 22.  1.300 */
 {  79, 0xD4 },  /* 23.  1.308 */
 {  85, 0x31 },  /* 24.  1.333 */
 {  91, 0xE5 },  /* 25.  1.357 */
 {  93, 0xB4 },  /* 26.  1.364 */
 {  96, 0x83 },  /* 27.  1.375 */
 {  98, 0xD5 },  /* 28.  1.385 */
 { 102, 0x52 },  /* 29.  1.400 */
 { 107, 0xC5 },  /* 30.  1.417 */
 { 110, 0x73 },  /* 31.  1.429 */
 { 114, 0x94 },  /* 32.  1.444 */
 { 116, 0xB5 },  /* 33.  1.455 */
 { 118, 0xD6 },  /* 34.  1.462 */
 { 119, 0xF7 },  /* 35.  1.467 */
 { 128, 0x21 },  /* 36.  1.500 */
 { 137, 0xF8 },  /* 37.  1.533 */
 { 138, 0xD7 },  /* 38.  1.538 */
 { 140, 0xB6 },  /* 39.  1.545 */
 { 142, 0x95 },  /* 40.  1.556 */
 { 146, 0x74 },  /* 41.  1.571 */
 { 149, 0xC7 },  /* 42.  1.583 */
 { 154, 0x53 },  /* 43.  1.600 */
 { 158, 0xD8 },  /* 44.  1.615 */
 { 160, 0x85 },  /* 45.  1.625 */
 { 163, 0xB7 },  /* 46.  1.636 */
 { 165, 0xE9 },  /* 47.  1.643 */
 { 171, 0x32 },  /* 48.  1.667 */
 { 177, 0xD9 },  /* 49.  1.692 */
 { 179, 0xA7 },  /* 50.  1.700 */
 { 183, 0x75 },  /* 51.  1.714 */
 { 186, 0xB8 },  /* 52.  1.727 */
 { 188, 0xFB },  /* 53.  1.733 */
 { 192, 0x43 },  /* 54.  1.750 */
 { 197, 0xDA },  /* 55.  1.769 */
 { 199, 0x97 },  /* 56.  1.778 */
 { 201, 0xEB },  /* 57.  1.786 */
 { 205, 0x54 },  /* 58.  1.800 */
 { 209, 0xB9 },  /* 59.  1.818 */
 { 213, 0x65 },  /* 60.  1.833 */
 { 217, 0xDB },  /* 61.  1.846 */
 { 219, 0x76 },  /* 62.  1.857 */
 { 222, 0xFD },  /* 63.  1.867 */
 { 224, 0x87 },  /* 64.  1.875 */
 { 228, 0x98 },  /* 65.  1.889 */
 { 230, 0xA9 },  /* 66.  1.900 */
 { 233, 0xBA },  /* 67.  1.909 */
 { 235, 0xCB },  /* 68.  1.917 */
 { 236, 0xDC },  /* 69.  1.923 */
 { 238, 0xED },  /* 70.  1.929 */
 { 239, 0xFE },  /* 71   1.933 */
};


/* UART Interrupt.
der Interrupt kommt nur beim LeerWERDEN,
nicht jedoch durch LeerSEIN.
Anstoen nur per Software-Interrupt!!
 */

__irq void UART0_IRQHandler (void)
{ long L;
  L = LPC_UART0->IIR & 1;                     /* 1 = kein Int gewesen */
  if (L) LPC_UART0->IER  = 2;                 /* war Soft-Int, also THRE enable */
  if (LPC_UART0->LSR & (1<<5))                /* wenn THRE = 1 */
  { if (v24sin0==v24sout0)  return;           /* raus wenn Puffer leer */
    LPC_UART0->THR = V24puffer0[v24sout0];    /* Zeichen senden      */
    v24sout0 = (v24sout0+1) & (v24bufsize-1); /* Index weiterstellen */
  }
}

__irq void UART1_IRQHandler (void)
{ long L;
  L = LPC_UART1->IIR & 1;                     /* 1 = kein Int gewesen */
  if (L) LPC_UART1->IER  = 2;                 /* war Soft-Int, also THRE enable */
  if (LPC_UART1->LSR & (1<<5))                /* wenn THRE = 1 */
  { if (v24sin1==v24sout1)  return;           /* raus wenn Puffer leer */
    LPC_UART1->THR = V24puffer1[v24sout1];    /* Zeichen senden      */
    v24sout1 = (v24sout1+1) & (v24bufsize-1); /* Index weiterstellen */
  }
}

__irq void UART2_IRQHandler (void)
{ long L;
  L = LPC_UART2->IIR & 1;                     /* 1 = kein Int gewesen */
  if (L) LPC_UART2->IER  = 2;                 /* war Soft-Int, also THRE enable */
  if (LPC_UART2->LSR & (1<<5))                /* wenn THRE = 1 */
  { if (v24sin2==v24sout2)  return;           /* raus wenn Puffer leer */
    LPC_UART2->THR = V24puffer2[v24sout2];    /* Zeichen senden      */
    v24sout2 = (v24sout2+1) & (v24bufsize-1); /* Index weiterstellen */
  }
}

__irq void UART3_IRQHandler (void)
{ long L;
  L = LPC_UART3->IIR & 1;                     /* 1 = kein Int gewesen */
  if (L) LPC_UART3->IER  = 2;                 /* war Soft-Int, also THRE enable */
  if (LPC_UART3->LSR & (1<<5))                /* wenn THRE = 1 */
  { if (v24sin3==v24sout3)  return;           /* raus wenn Puffer leer */
    LPC_UART3->THR = V24puffer3[v24sout3];    /* Zeichen senden      */
    v24sout3 = (v24sout3+1) & (v24bufsize-1); /* Index weiterstellen */
  }
}



/* UART  initialisieren */
/* ==================== */
/*
 Baudraten-Rechengang fr alle LPC2xxx, auch LPC17xx:
     Divisor = UART0_CLK / (Baudrate * K);

 ohne fraktionalen Vorteiler ist K = 16
 mit ihm kann K sein: 16.00 bis 30.03 in 72 Stufen
 Es gibt DIVADVAL (0..15) und MULVAL (1..15), wobei gilt:
     K = 16 * (1 + DIVADVAL/MULVAL));
 */

dword InitSerial (long baudrate, word wo)
{ long divisor;
  long rest;
  long ksoll;
  long kist;
  long brcal;
  byte FDRval;
  int  i;
  dword CLOCK;

  switch (wo)
  { case toUART0:
             PCONP |= (1<<3);
             LPC_UART0->LCR = 0x83;    /* 8 bits, no Parity, 1 Stop bit     */
             CLOCK = UART0Clock;
	     break;

    case toUART1:
             PCONP |= (1<<4);
             LPC_UART1->LCR = 0x83;    /* 8 bits, no Parity, 1 Stop bit     */
             CLOCK = UART1Clock;
	     break;

    case toUART2:
             PCONP |= (1<<24);
             LPC_UART2->LCR = 0x83;    /* 8 bits, no Parity, 1 Stop bit     */
             CLOCK = UART2Clock;
	     break;

    case toUART3:
             PCONP |= (1<<25);
             LPC_UART3->LCR = 0x83;    /* 8 bits, no Parity, 1 Stop bit     */
             CLOCK = UART3Clock;
	     break;

    default: CLOCK = UART0Clock;
  }

  divisor = CLOCK / baudrate;                /* erstmal das 16 fache */
  rest = divisor & 15;                       /* Rest merken */
  divisor = divisor >> 4;
  FDRval = (1<<4);                           /* Fraktionalteiler aus       */

  if (rest > 13)                             /* sind nahe genug an integer */
  { ++divisor;                               /* divisor aufrunden */
    brcal = CLOCK / (16 * divisor);          /* Returnwert */
    goto _setzebaud;
  }

  if (rest < 2)                              /* sind nahe genug an integer */
  { brcal = CLOCK / (16 * divisor);          /* Returnwert */
    goto _setzebaud;
  }


  /* ok, wir sind in der Prrie und suchen jetzt divisor, MULVAL und DIVADVAL,
     merke: teiler = PCLK / (baudrate * K);  K von 16 bis ca. 31,
     wir nehmen die Mitte, also 75% vom eigentlichen teiler
   */
  divisor = divisor - (divisor>>2);          /* Divisor = 0.75 * divisor;  */

  /* der divisor steht erstmal, jetzt zurckrechnen, um K zu bestimmen */
  ksoll = CLOCK * 16;
  ksoll = ksoll / (baudrate * divisor);      /* ksoll ist jetzt (16 .. ca. 31) * 16 */
  ksoll = ksoll - 255;                       /* um nicht einseitig zu suchen   */

  /* passenden Wert in der Tabelle suchen */
  kist = 2048;
  i = 72;
  while (kist > ksoll)
  { --i;
    kist = FracTabelle[i].div;
    FDRval = FracTabelle[i].FDR;
  }

  /* Returnwert berechnen (in long!!) */
  kist = kist + 256;
  brcal = CLOCK * 16;
  brcal = brcal / (divisor * kist);

  _setzebaud:
  switch (wo)
  { case toUART0:
             LPC_UART0->FDR = FDRval;                /* Fraktionalteiler            */
             LPC_UART0->DLL = divisor & 0xFF;        /* lo Teil des Divisors        */
             LPC_UART0->DLM = (divisor >> 8) & 0xFF; /* hi Teil des Divisors        */
             LPC_UART0->FCR = 7;                     /* Enable FIFO und Rcksetzen  */
             LPC_UART0->LCR = 0x03;                  /* DLAB = 0                    */
             v24sin0 = v24sout0 = 0;                 /* TX Puffer lschen           */
             LPC_UART0->FCR = 1;                     /* Enable Fifo                 */
             NVIC_ISER0 = 1<<5;                      /* UART 0 hat Bit #5           */
             LPC_UART0->IER = 2;                     /* enable TX-Interrupt         */
	     break;

    case toUART1:
             LPC_UART1->FDR = FDRval;                /* Fraktionalteiler            */
             LPC_UART1->DLL = divisor & 0xFF;        /* lo Teil des Divisors        */
             LPC_UART1->DLM = (divisor >> 8) & 0xFF; /* hi Teil des Divisors        */
             LPC_UART1->FCR = 7;                     /* Enable FIFO und Rcksetzen  */
             LPC_UART1->LCR = 0x03;                  /* DLAB = 0                    */
             v24sin1 = v24sout1 = 0;                 /* TX Puffer lschen           */
             LPC_UART1->FCR = 1;                     /* Enable Fifo                 */
             NVIC_ISER0 = 1<<6;                      /* UART 1 hat Bit #6           */
             LPC_UART1->IER = 2;                     /* enable TX-Interrupt         */
	     break;

    case toUART2:
             LPC_UART2->FDR = FDRval;                /* Fraktionalteiler            */
             LPC_UART2->DLL = divisor & 0xFF;        /* lo Teil des Divisors        */
             LPC_UART2->DLM = (divisor >> 8) & 0xFF; /* hi Teil des Divisors        */
             LPC_UART2->FCR = 7;                     /* Enable FIFO und Rcksetzen  */
             LPC_UART2->LCR = 0x03;                  /* DLAB = 0                    */
             v24sin2 = v24sout2 = 0;                 /* TX Puffer lschen           */
             LPC_UART2->FCR = 1;                     /* Enable Fifo                 */
             NVIC_ISER0 = 1<<7;                      /* UART 2 hat Bit #7           */
             LPC_UART2->IER = 2;                     /* enable TX-Interrupt         */
	     break;

    case toUART3:
             LPC_UART3->FDR = FDRval;                /* Fraktionalteiler            */
             LPC_UART3->DLL = divisor & 0xFF;        /* lo Teil des Divisors        */
             LPC_UART3->DLM = (divisor >> 8) & 0xFF; /* hi Teil des Divisors        */
             LPC_UART3->FCR = 7;                     /* Enable FIFO und Rcksetzen  */
             LPC_UART3->LCR = 0x03;                  /* DLAB = 0                    */
             v24sin3 = v24sout3 = 0;                 /* TX Puffer lschen           */
             LPC_UART3->FCR = 1;                     /* Enable Fifo                 */
             NVIC_ISER0 = 1<<8;                      /* UART 3 hat Bit #8           */
             LPC_UART3->IER = 2;                     /* enable TX-Interrupt         */
	     break;
  }
  return brcal;
}


/* Zeichen zur V24 ausgeben.
   =========================
Die Zeichen werden immer nur in den Sendepuffer geschrieben,
nie direkt in die Hardware - das macht der Interrupt.
Zum eventuellen Starten wird ein Int per Software ausgelst
Danach wird gewartet, bis wenigstens 1 Platz frei ist im
Sendepuffer (mit Timeout, damit er nicht hngen bleibt).
Rckgabe: c oder 0 bei Timeout
*/

char V24Char_Out0 (char c)
{ int  h, n;
  long L;

  h = v24sin0;
  n = (h + 1) & (v24bufsize-1);            /* h = nchster Ringpufferplatz   */
  L = 2000;                                /* wir knnen nicht ewig warten!  */
  while (n==v24sout0)
  { NVIC_ISPR0 = (1<<5);
    delay(1000);
    --L;  if (!L) return 0;
  }
  V24puffer0[h] = c;
  v24sin0 = n;
  NVIC_ISPR0 = (1<<5);                     /* Int auslsen zum evtl. Starten */
  return c;
}

char V24Char_Out1 (char c)
{ int  h, n;
  long L;

  h = v24sin1;
  n = (h + 1) & (v24bufsize-1);            /* h = nchster Ringpufferplatz   */
  L = 2000;                                /* wir knnen nicht ewig warten!  */
  while (n==v24sout1)
  { NVIC_ISPR0 = (1<<6);
    delay(1000);
    --L;  if (!L) return 0;
  }
  V24puffer1[h] = c;
  v24sin1 = n;
  NVIC_ISPR0 = (1<<6);                     /* Int auslsen zum evtl. Starten */
  return c;
}

char V24Char_Out2 (char c)
{ int  h, n;
  long L;

  h = v24sin2;
  n = (h + 1) & (v24bufsize-1);            /* h = nchster Ringpufferplatz   */
  L = 2000;                                /* wir knnen nicht ewig warten!  */
  while (n==v24sout2)
  { NVIC_ISPR0 = (1<<7);
    delay(1000);
    --L;  if (!L) return 0;
  }
  V24puffer2[h] = c;
  v24sin2 = n;
  NVIC_ISPR0 = (1<<7);                     /* Int auslsen zum evtl. Starten */
  return c;
}

char V24Char_Out3 (char c)
{ int  h, n;
  long L;

  h = v24sin3;
  n = (h + 1) & (v24bufsize-1);            /* h = nchster Ringpufferplatz   */
  L = 2000;                                /* wir knnen nicht ewig warten!  */
  while (n==v24sout3)
  { NVIC_ISPR0 = (1<<8);
    delay(1000);
    --L;  if (!L) return 0;
  }
  V24puffer3[h] = c;
  v24sin3 = n;
  NVIC_ISPR0 = (1<<8);                     /* Int auslsen zum evtl. Starten */
  return c;
}


/* Anzahl freier Pltze im Sendepuffer ermitteln
   =============================================
 */
int V24numTxFree0 (void)
{ int i;
  i = v24sin0 - v24sout0;    /* i = belegte Pltze */
  if (i<0) i = i + v24bufsize;
  return v24bufsize - i;
}

int V24numTxFree1 (void)
{ int i;
  i = v24sin1 - v24sout1;    /* i = belegte Pltze */
  if (i<0) i = i + v24bufsize;
  return v24bufsize - i;
}

int V24numTxFree2 (void)
{ int i;
  i = v24sin2 - v24sout2;    /* i = belegte Pltze */
  if (i<0) i = i + v24bufsize;
  return v24bufsize - i;
}

int V24numTxFree3 (void)
{ int i;
  i = v24sin3 - v24sout3;    /* i = belegte Pltze */
  if (i<0) i = i + v24bufsize;
  return v24bufsize - i;
}


/* Test, ob Zeichen an USART da ist
   ================================
 */
bool V24RxAvail0 (void)
{ if (LPC_UART0->LSR & 1) return 1; else return 0; }

bool V24RxAvail1 (void)
{ if (LPC_UART1->LSR & 1) return 1; else return 0; }

bool V24RxAvail2 (void)
{ if (LPC_UART2->LSR & 1) return 1; else return 0; }

bool V24RxAvail3 (void)
{ if (LPC_UART3->LSR & 1) return 1; else return 0; }


/* Zeichen auf USART empfangen (ohne Echo)
   =======================================
 */
char V24GetChar0 (void)
{ if (!(LPC_UART0->LSR & 1)) return 0;  return (LPC_UART0->RBR); }

char V24GetChar1 (void)
{ if (!(LPC_UART1->LSR & 1)) return 0;  return (LPC_UART1->RBR); }

char V24GetChar2 (void)
{ if (!(LPC_UART2->LSR & 1)) return 0;  return (LPC_UART2->RBR); }

char V24GetChar3 (void)
{ if (!(LPC_UART3->LSR & 1)) return 0;  return (LPC_UART3->RBR); }


/* Warten, bis alle Zeichen gesendet sind
   ======================================
 */
void q (void)
{ FIO2SET = 1; FIO2CLR =  1; }

void V24TxDone0 (void) { while (v24sin0!=v24sout0) q(); }
void V24TxDone1 (void) { while (v24sin1!=v24sout1) q(); }
void V24TxDone2 (void) { while (v24sin2!=v24sout2) q(); }
void V24TxDone3 (void) { while (v24sin3!=v24sout3) q(); }



/*  Ende  */
